#include <u.h>
#include <libc.h>
#include <auth.h>
#include <bio.h>
#include <libsec.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

void authdig1(char *user, char *realm, char *passwd, char *ha1);
void authdig2(char *ha1, char *nonce, char *method, char *uri, char *dig);

static	void		printtype(Hio *hout, HContent *type, HContent *enc);
static	int	md5fmt(Fmt *f);
static	char*	http_auth(HConnect *c, char *file, char *path);
int		putnonce(char *nonce, char *res);
void		rmvnonce(char *);
int			chknonce(char *nonce, char *res);
char		*gennonce(void);

/*
 * check for authorization for some parts of the server tree
 * the user name supplied with the authorization request is ignored;
 * instead, we authenticate as the realm's user.
 *
 * authorization should be done before opening any files so that
 * unauthorized users don't get to validate file names.
 *
 * returns 1 if authorized, 0 if unauthorized, -1 for io failure.
 */
int
authcheck(HConnect *c, char *pwfile, char *path)
{
	Hio *hout;
	int status;
	char *realm, *pp;
	char *nonce;
	int i;
	char *s,*e;
	char buf[256];
	int isbasic = 0;

	realm = http_auth(c, pwfile, path);
	/* 	this realm must be passed to the client so that client
	*	can show me the credential for this realm.
	*	-Kenar-	*/
	if(realm == nil)
		return 1;
	pp = nil;
	if(*realm == '"'){ /* then the real is for basic credential. */
		realm++;
		isbasic = 1;
	}
	else{ /* then return value is: realm + '"' + pp */
		s = strchr(realm,'"');
		*s++ = 0;
		pp = s;
	}

	if(!isbasic){
		char *stale;
		if(c->head.nextnonce)
			nonce = strdup(c->head.nextnonce);
		else{
			nonce = gennonce();
			putnonce(nonce,nil); /* register the nonce value */
		}
		stale = c->head.stale?"true":"false";
		/* 	stale == true	the client used old once value.
		*	stale != true	the client will prompted to input authentication date
		*	see rfc2069 or rfc2617 for more details */
		snprint(buf, sizeof(buf),
			"Digest realm=\"%s\", domain=\"%s\", nonce=\"%s\", stale=%s, qop=\"auth\"",
			realm,pp,nonce,stale);
		db_put(&errh, "WWW-Authenticate:", buf);
		//logit(c,"DEBUG: %s", buf);
		free(nonce);
		c->head.nextnonce = nil;
		c->head.stale = 0;
	}

	if(allowbasic){
		snprint(buf, sizeof(buf), "Basic realm=\"%s\"", realm);
		db_add(&errh, "WWW-Authenticate:", buf);
	}
	status = hfail(c, HUnauth, path);
	/* 0: success; -1: failure */
	if(status < 0)
		logit(c, "%s authcheck: hflush.", c->req.uri);
	/*	Then the browser will show a login window to the user
	*	and will resend the request with some authentication
	*	information.
	*	Safari 2.0.3 is something weird. It sometimes doesn't show the login window.
	*	-Kenar- */
	return status;
}


/*
 * fd references a file which has been authorized & checked for relocations.
 * send back the headers & it's contents.
 * includes checks for conditional requests & ranges.
 */
int
sendit(HConnect *c, char *file, int hctl)
{
	HRange *r;
	HContents conts;
	HContent *enc=nil;
	HContent *type=nil;
	Hio *hout;
	char *boundary, etag[32];
	int fd;
	Dir *dir;
	ulong tr;
	int n, nw, multir, ok;
	vlong wrote;
	char *s, *p;

	hout = &c->hout;

	/*	content type will be determined as following
	*	case CGI
	*	1. Explicit declaration in "Content-Type:" in CGI
	*	2. default value in /etc/handler
	*	case regular file
	*	1. URI, i.e., suffix of the file
	*	2. the content of the requested data
	*	-Kenar-	*/

	s = db_get(cgih, "Content-Type:");
	if(s){
		s = hstrdup(c,s);
		p = strchr(s,'/');
		if(p == nil)
			return hfail(c, HInternal);
		*p = 0;
		type = halloc(c,sizeof(HContent));
		type->generic = s;
		type->specific = ++p;
		db_del(&cgih, "Content-Type:");
	}

	/*	content encoding will be determined as following
	*	1. Explicit declaration in "Content-Encoding:" in CGI
	*	2. URI, i.e., suffix of the requested file
	*	3. the content of the requested data
	*	-Kenar-	*/

	s = db_get(cgih, "Content-Encoding:");
	if(s){
		enc = hmkcontent(c, s, nil, nil);
		db_del(&cgih, "Content-Encoding:");
	}

	fd = open(file, OREAD);
	if(fd < 0){
		char buf[256];

		if(strcmp(file,"/tmp/...") == 0){ // Kenar
			logit(c, "/tmp/... not open");
			return hfail(c, HInternal);
		}
		file = file + 4; // 4 is strlen("/doc")
		logit(c, "sendit: %s not found", file);
		buf[0] = 0;
		errstr(buf, sizeof buf);
		if(strstr(buf, "permission denied"))
			return hfail(c, HForbidden, file);
		return hfail(c, HNotFound, file);
	}
	if((dir = dirfstat(fd)) == nil){
		close(fd);
		logit(c, "# error: sendit: dirfstat: %s", file);
		return hfail(c, HInternal);
	}

	/*
	 * figure out the type of file and send headers
	 */
	n = -1;
	r = nil;
	multir = 0;
	boundary = nil;
	if(c->req.vermaj){
		if(type == nil && enc == nil){
			conts = uriclass(c, c->req.path);// Kenar
			//conts = uriclass(c, c->req.uri);// original code
			type = conts.type;
			enc = conts.encoding;
			if(type == nil && enc == nil){
				n = read(fd, c->xferbuf, HBufSize-1);
				if(n > 0){
					c->xferbuf[n] = '\0';
					conts = dataclass(c, c->xferbuf, n);
					type = conts.type;
					enc = conts.encoding;
				}
			}
		}
		if(type == nil)
			type = hmkcontent(c, "application", "octet-stream", nil);

		snprint(etag, sizeof(etag), "\"%luxv%lux\"", dir->qid.path, dir->qid.vers);
		ok = checkreq(c, type, enc, dir->mtime, etag);
		if(ok <= 0){
			close(fd);
			return ok;
		}

		/*
		 * check for if-range requests
		 */
		if(c->head.range == nil
		|| c->head.ifrangeetag != nil && !etagmatch(1, c->head.ifrangeetag, etag)
		|| c->head.ifrangedate != 0 && c->head.ifrangedate != dir->mtime){
			c->head.range = nil;
			c->head.ifrangeetag = nil;
			c->head.ifrangedate = 0;
		}

		if(c->head.range != nil){
			c->head.range = fixrange(c->head.range, dir->length);
			if(c->head.range == nil){
				if(c->head.ifrangeetag == nil && c->head.ifrangedate == 0){
					char buf[64];
					snprint(buf, sizeof(buf), "bytes */%lld", dir->length);
					db_put(&errh, "Content-Range:", buf);
					hfail(c, HReqRange);
					return 1;
				}
				c->head.ifrangeetag = nil;
				c->head.ifrangedate = 0;
			}
		}

		if(c->head.range == nil){
			s = db_get(cgih, "Status:");
			if(s == nil)
				s = "200 OK";
			hprint(hout, "%s %s\r\n", hversion, s);
			db_del(&cgih,"Status:");
		}
		else
			hprint(hout, "%s 206 Partial Content\r\n", hversion);

		s = db_get(cgih, "Server:");
		if(s == nil)
			s = serversoft;
		hprint(hout, "Server: %s\r\n", s);
		db_del(&cgih,"Server:");

		s = db_get(cgih, "Date:");
		if(s)
			hprint(hout, "Date: %s\r\n", s);
		else
			hprint(hout, "Date: %D\r\n", time(nil));
		db_del(&cgih,"Date:");

		s = db_get(cgih, "ETag:");
		if(s)
			hprint(hout, "ETag: %s\r\n", s);
		else
			hprint(hout, "ETag: %s\r\n", etag);
		db_del(&cgih,"ETag:");

		hprint(hout, "Accept-Ranges: bytes\r\n");

		/*
		 * can't send some entity headers if partially responding
		 * to an if-range: etag request
		 */
		r = c->head.range;
		if(r == nil){
			hprint(hout, "Content-Length: %lld\r\n", dir->length);
		}
		else if(r->next == nil){
			hprint(hout, "Content-Range: bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
			logit(c, "Content-Range:  bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
		}else{
			multir = 1;
			boundary = hmkmimeboundary(c);
			hprint(hout, "Content-Type: multipart/byteranges; boundary=%s\r\n", boundary);
		}
		if(c->head.ifrangeetag == nil){
			hprint(hout, "Last-Modified: %D\r\n", dir->mtime);
			if(!multir)
				printtype(hout, type, enc);
			if(c->head.fresh_thresh)
				hintprint(c, hout, c->req.uri, c->head.fresh_thresh, c->head.fresh_have);
		}
		if(strcmp(file,"/tmp/...") == 0) // Kenar
			hprint(hout, "Cache-Control: no-cache\r\n");

		if(c->head.closeit)
			hprint(hout, "Connection: close\r\n");
		else if(!http11(c))
			hprint(hout, "Connection: Keep-Alive\r\n");

#ifdef nodef
		if(c->head.nextnonce){
			/* 	support nextnonce of RFC2069 clients.
			*	but it seems no working
			*	-Kenar- */
			hprint(hout, "Authentication-Info: nextnonce=\"%s\"\r\n",c->head.nextnonce);
			//logit(c,"DEBUG: sendit: nextnonce=%s is attached", c->head.nextnonce);
			free(c->head.nextnonce);
			c->head.nextnonce = nil;
		}
#endif

		/* If hctl == HCPlus then add the headers" */
		if(hctl == HCPlus){
			Db *d;
			d = cgih;
			while(d){
				hprint(hout,"%s %s\r\n", d->key, d->val);
				d = d->next;
			}
		}
		hprint(hout, "\r\n");
	}
	if(strcmp(c->req.meth, "HEAD") == 0){
		if(c->head.range == nil)
			logit(c, "Reply: 200 file 0");
		else
			logit(c, "Reply: 206 file 0");
		hflush(hout);
		close(fd);
		return 1;
	}

	/*
	 * send the file if it's a normal file
	 */
	if(r == nil){
		hflush(hout);

		wrote = 0;
		if(n > 0)
			wrote = write(1, c->xferbuf, n);
		if(n <= 0 || wrote == n){
			while((n = read(fd, c->xferbuf, HBufSize)) > 0){
				nw = write(1, c->xferbuf, n);
				if(nw != n){
					if(nw > 0)
						wrote += nw;
					break;
				}
				wrote += nw;
			}
		}
		close(fd);
		if(dir->length == wrote){
			if(strcmp(file, "/tmp/...") != 0) // Kenar
				logit(c,"sent %s", file);
			return 1;
		}
		logit(c, "Reply: 200 file %lld %lld", dir->length, wrote);
		return -1;
	}

	/*
	 * for multipart/byterange messages,
	 * it is not ok for the boundary string to appear within a message part.
	 * however, it probably doesn't matter, since there are lengths for every part.
	 */
	wrote = 0;
	ok = 1;
	for(; r != nil; r = r->next){
		if(multir){
			hprint(hout, "\r\n--%s\r\n", boundary);
			printtype(hout, type, enc);
			hprint(hout, "Content-Range: bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
			hprint(hout, "\r\n");
		}
		hflush(hout);

		if(seek(fd, r->start, 0) != r->start){
			ok = -1;
			break;
		}
		for(tr = r->stop - r->start + 1; tr; tr -= n){
			n = tr;
			if(n > HBufSize)
				n = HBufSize;
			if(read(fd, c->xferbuf, n) != n){
				ok = -1;
				goto breakout;
			}
			nw = write(1, c->xferbuf, n);
			if(nw != n){
				if(nw > 0)
					wrote += nw;
				ok = -1;
				goto breakout;
			}
			wrote += nw;
		}
	}
breakout:;
	if(r == nil){
		if(multir){
			hprint(hout, "--%s--\r\n", boundary);
			hflush(hout);
		}
		logit(c, "Reply: 206 partial content %lld %lld", dir->length, wrote);
	}else
		logit(c, "Reply: 206 partial content, early termination %lld %lld", dir->length, wrote);
	close(fd);
	return ok;
}

static void
printtype(Hio *hout, HContent *type, HContent *enc)
{
	hprint(hout, "Content-Type: %s/%s", type->generic, type->specific);
/*
	if(cistrcmp(type->generic, "text") == 0)
		hprint(hout, ";charset=utf-8");
*/
	hprint(hout, "\r\n");
	if(enc != nil)
		hprint(hout, "Content-Encoding: %s\r\n", enc->generic);
}

int
etagmatch(int strong, HETag *tags, char *e)
{
	char *s, *t;

	for(; tags != nil; tags = tags->next){
		if(strong && tags->weak)
			continue;
		s = tags->etag;
		if(s[0] == '*' && s[1] == '\0')
			return 1;

		t = e + 1;
		while(*t != '"'){
			if(*s != *t)
				break;
			s++;
			t++;
		}

		if(*s == '\0' && *t == '"')
			return 1;
	}
	return 0;
}

static char *
acceptcont(char *s, char *e, HContent *ok, char *which)
{
	char *sep;

	if(ok == nil)
		return seprint(s, e, "Your browser accepts any %s.<br>\n", which);
	s = seprint(s, e, "Your browser accepts %s: ", which);
	sep = "";
	for(; ok != nil; ok = ok->next){
		if(ok->specific)
			s = seprint(s, e, "%s%s/%s", sep, ok->generic, ok->specific);
		else
			s = seprint(s, e, "%s%s", sep, ok->generic);
		sep = ", ";
	}
	return seprint(s, e, ".<br>\n");
}

/*
 * send back a nice error message if the content is unacceptable
 * to get this message in ie, go to tools, internet options, advanced,
 * and turn off Show Friendly HTTP Error Messages under the Browsing category
 */
static int
notaccept(HConnect *c, HContent *type, HContent *enc, char *which)
{
	Hio *hout;
	char *s, *e;

	hout = &c->hout;
	e = &c->xferbuf[HBufSize];
	s = c->xferbuf;
	s = seprint(s, e, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
	s = seprint(s, e, "<html>\n<title>Unacceptable %s</title>\n<body>\n", which);
	s = seprint(s, e, "Your browser will not accept this data, %H, because of it's %s.<br>\n", c->req.uri, which);
	s = seprint(s, e, "It's Content-Type is %s/%s", type->generic, type->specific);
	if(enc != nil)
		s = seprint(s, e, ", and Content-Encoding is %s", enc->generic);
	s = seprint(s, e, ".<br>\n\n");

	s = acceptcont(s, e, c->head.oktype, "Content-Type");
	s = acceptcont(s, e, c->head.okencode, "Content-Encoding");
	s = seprint(s, e, "</body>\n</html>\n");
	return hfail0(c,"406 Not Acceptable", "Not Acceptable", s);
}

/*
 * check time and entity tag conditions.
 */
int
checkreq(HConnect *c, HContent *type, HContent *enc, long mtime, char *etag)
{
	Hio *hout;
	int m;

	hout = &c->hout;
	if(c->req.vermaj >= 1 && c->req.vermin >= 1 && !hcheckcontent(type, c->head.oktype, "Content-Type", 0))
		return notaccept(c, type, enc, "Content-Type");
	if(c->req.vermaj >= 1 && c->req.vermin >= 1 && !hcheckcontent(enc, c->head.okencode, "Content-Encoding", 0))
		return notaccept(c, type, enc, "Content-Encoding");

	/*
	 * can use weak match only with get or head;
	 * this always uses strong matches
	 */
	m = etagmatch(1, c->head.ifnomatch, etag);

	if(m && strcmp(c->req.meth, "GET") != 0 && strcmp(c->req.meth, "HEAD") != 0
	|| c->head.ifunmodsince && c->head.ifunmodsince < mtime
	|| c->head.ifmatch != nil && !etagmatch(1, c->head.ifmatch, etag))
		return hfail(c, HPrecond);

	if(c->head.ifmodsince >= mtime
	&& (m || c->head.ifnomatch == nil)){
		/*
		 * can only send back Date, ETag, Content-Location,
		 * Expires, Cache-Control, and Vary entity-headers
		 */
		hprint(hout, "%s 304 Not Modified\r\n", hversion);
		hprint(hout, "Server: %s\r\n", serversoft);
		hprint(hout, "Date: %D\r\n", time(nil));
		hprint(hout, "ETag: %s\r\n", etag);
		if(c->head.closeit)
			hprint(hout, "Connection: close\r\n");
		else if(!http11(c))
			hprint(hout, "Connection: Keep-Alive\r\n");
		hprint(hout, "\r\n");
		logit(c, "Reply: 304 Not Modified");
		return hflush(hout);
	}
	return 1;
}

/*
 * length is the actual length of the entity requested.
 * discard any range requests which are invalid,
 * ie start after the end, or have stop before start.
 * rewrite suffix requests
 */
HRange*
fixrange(HRange *h, long length)
{
	HRange *r, *rr;

	if(length == 0)
		return nil;

	/*
	 * rewrite each range to reflect the actual length of the file
	 * toss out any invalid ranges
	 */
	rr = nil;
	for(r = h; r != nil; r = r->next){
		if(r->suffix){
			r->start = length - r->stop;
			if(r->start >= length)
				r->start = 0;
			r->stop = length - 1;
			r->suffix = 0;
		}
		if(r->stop >= length)
			r->stop = length - 1;
		if(r->start > r->stop){
			if(rr == nil)
				h = r->next;
			else
				rr->next = r->next;
		}else
			rr = r;
	}

	/*
	 * merge consecutive overlapping or abutting ranges
	 *
	 * not clear from rfc2616 how much merging needs to be done.
	 * this code merges only if a range is adjacent to a later starting,
	 * over overlapping or abutting range.  this allows a client
	 * to request wanted data first, followed by other data.
	 * this may be useful then fetching part of a page, then the adjacent regions.
	 */
	if(h == nil)
		return h;
	r = h;
	for(;;){
		rr = r->next;
		if(rr == nil)
			break;
		if(r->start <= rr->start && r->stop + 1 >= rr->start){
			if(r->stop < rr->stop)
				r->stop = rr->stop;
			r->next = rr->next;
		}else
			r = rr;
	}
	return h;
}

/*  ----------- the followings are my addition -----Kenar---- */

static int
md5fmt(Fmt *f)
{
	char buf[MD5dlen*2+1];
	uchar *p;
	int i;

	p = va_arg(f->args, uchar*);
	for(i=0; i<MD5dlen; i++)
		sprint(buf+2*i, "%.2ux", p[i]);
//	strconv(buf, f);
	return fmtprint(f, "%s", buf);
}


/*
*	local authentication based on HTTP access authentication (RFC2617)
*
*	return values:
*	nil: OK
*	realm: NG at realm
*
*	password file is: /etc/passwd
*	the format is:
*		user md5sum apath
*	or
*		user passwd apath real
*	where apath is the pathname that should be authenticated.
*
*	The first format is provide for compatibility to Pegasus 2.1
*	that support only basic authentication.
*	The second format is for both digest and basic authentication
*	that is provided for Pegasus 2.2
*
*	Example for old format:
*		alice		54ef36ec71201fdf9d1423fd26f97f6b	/private
*	In getting the second field, be careful not to include trailing '\n'.
*	Therefore execute like this:
*	echo -n 'open sesame' | md5sum
*
*	New format contains 4 fields like:
*		alice		'open sesame'	/private //hera/~alice/private
*
*	The passwd file must be like the following:
*	--rw-r----- M 8 arisawa web 323 Jun 22  2006 passwd
*
*	Return value is realm if the path is protected.
*	If not, return value is nil.
*	Caller must know whether the protection is basic or digest.
*	How to do ?
*	We can make use of double quotation to mark a realm value
*	for basic authethentication, because double quotation is
*	prohibitted in realm name.
*
*	If return value begins with '"' then we must remove the symbol
*	and we can judge the real is for basic authentication.
*	This trick is for compatibility of previous version format of passwd
*	file.
*
*	-Kenar-	*/
char*
http_auth(HConnect *c, char *pwfile, char *path)
/*
	c->head.authuser
	c->head.authpass
	c->head.authresponse
	c->head.authnonce
	c->head.authrealm
	...

	(nonce,response) is
		(nil,nil): forst time
		(nil,passwd): response in basic authentication
		(nonce,response): response in digest authentication

	response is a password in case of basic authentication
*/
{
	Biobuf *b;
	char *line, *s, *hash, *usr, *huri;
	char *field[5];
	uchar digest[MD5dlen];
	char dig[2*MD5dlen+1];
	DigestState *ds;
	char buf[128]; // 128 > 2*MD5dlen + 1
	static int fmt=0;
	char ch;
	int ok, old;
	int m, m0, n;
	char *realm, *pp, *ret;
	char *user, *passwd, *response, *nonce, *method, *uri, *ouri;
	char *nc, *cnonce, *qop;
	char *nextnonce=nil;
	static Db *arealm; /* authenticated realms */

	user = c->head.authuser;
	huri = c->head.authuri; /* used in digest authentication hash */
	passwd = c->head.authpass;
	nonce = c->head.authnonce;
	response = c->head.authresponse;
	nc = c->head.authnc;
	cnonce = c->head.authcnonce;
	qop = c->head.authqop;
	
	uri = c->req.uri;
	ouri = c->req.ouri;
	method = c->req.meth;

	b = Bopen(pwfile, OREAD);
	if(b == nil)
		return nil;

	/*	we should allow every printable character in password
	*	but currently not. this should be fixed. */
	m0 = 0;
	usr = nil;
	hash = nil;
	realm = nil;
	pp = nil;
	while((line = Brdline(b, '\n')) != nil){
		line[Blinelen(b)-1] = 0;
		/*	line that follow '#' is a comment */
		s = strchr(line, '#');
		if(s != nil)
			*s = '\0'; 	/* chop comment */
		n = tokenize(line, field, 5);
		if(n < 3)
			continue;
		/* then n==3 or n==4, field[2] is a protected path */
		m = deeper(path, field[2]);
		if(m > m0){
			m0 = m;
			usr = field[0];
			hash = field[1]; /* if n==3 then hash is md5sum of the passwd */
			pp = field[2]; /* protected path */
			if(n == 4)
				realm = field[3]; /* realm */
			else
				realm = nil; /* clean the stale realm */
		}
	}
	Bterm(b);
	//logit(c,"DEBUG: http_auth: usr=%s hash=%s pp=%s realm=%s",usr,hash,pp,realm);
	if(!pp) /* the path is not protected */
		return nil;

	old = 0;
	if(realm == nil){
		realm = pp;
		old = 1;
	}

	/*  If the realm is already authenticated, we may let
	*	the browser enter the realm even if no authentication info in the header
	*	This  convention is for Safari 2.0.3, but might work well for all browsers.
	*	This convention makes no security problem as long as connection is not hijacked.
	*	However a hostile proxy can compromise the security.
	*	Note: Safari thinks the URI
	*		/private/foo;param?query
	*	is out side of
	*		/private
	*/
	if(user == nil && db_get(arealm,realm)){
		//logit(c,"DEBUG: http_auth: db_get: realm=%s OK", realm);
		return nil;
	}

	/* we add "/~alice" to pp */
	if(c->req.user){
		m = strlen(c->req.user) + strlen(pp) + 10;
		s = pp;
		pp = halloc(c, m);
		snprint(pp,m, "/~%s%s", c->req.user, s);
	}
	
	/* the path is protected */
	/* ret is a return value */
	if(old){ /* old format. put tricky '"' */
		m = strlen(realm) + 5;
		ret = halloc(c, m);
		snprint(ret,m, "\"%s", realm);
		/* we must remove '"' later */
	}
	else{
		/* 	We must return both realm and pp
		*	How to do? We again use '"' as a separator,i.e.,
		*	We return realm + '"' + pp
		*	we must decompose to (realm,pp) later
		*	-Kenar- */
		m = strlen(realm) + strlen(pp) + 10;
		ret = halloc(c, m);
		snprint(ret,m, "%s\"%s", realm,pp);
	}
	/* then, if *ret is '"', hash is md5sum of password */

	//logit(c,"DEBUG: http_auth: usr=%s, user=%s ret=%s",usr,user,ret);

	ok = 0;
	if(!user || strcmp(user, usr) != 0)
		goto R;
	if(!passwd && !response)
		goto R;

	/* the client presented response or password */
	if(allowbasic && old){
		/* old format: hash is md5sum of password */
		ds = md5(nil, 0, nil, nil);
		md5((uchar*)passwd, strlen(passwd), digest, ds);
		if(fmt == 0){
			fmtinstall('M', md5fmt);
			fmt = 1;
		}
		snprint(buf, sizeof buf, "%M", digest);
		if(strcmp(hash, buf) == 0)
			ok = 1;
		//logit(c,"DEBUG: http_auth: Basic: %s %s %s",passwd,buf,hash);
	}
	else if(allowbasic && response == nil){
		/* then hash is md5sum of user:realm:passwd */
		char ha1[MD5dlen*2+1];
		authdig1(user,realm,passwd,ha1);
		if(strcmp(ha1,hash) == 0)
			ok = 1;
	}
	else { /* then digest authentication */
		char buf[128];
		char *xnonce;
		/*	we use a convention:
		*		if cnonce==nil then RFC2069 digest authentication
		*		else then RFC2617 digest authentication
		*
		*	we will assign:
		*	xnonce = nonce for RFC2069
		*	xnonce = nonce + ":" + nc + ":" + cnonce + ":" + qop  for RFC2617
		*
		* 	xnonce is used for RFC2617 support. the size is probably sufficient.
		* 	note: RFC2069 and RFC2617 does not say the length of nonce value.
		*	strlen(nc) is 8. (RFC2617)
		*	strlen(nonce) is determined by server
		*	strlen(cnonce) is determined by client
		*	Safari uses 32 LHEX for cnonce that is same as response
		*	-Kenar- */

		//logit(c,"DEBUG: http_auth: Digest: method=%s huri=%s",method,huri);
		xnonce = nonce;
		if(cnonce){ /* RFC2617 */
			n = snprint(buf,sizeof buf,"%s:%s:%s:%s",nonce,nc,cnonce,qop);
			/* snprint() returns strlen(buf), we examine the size */
			if(n >= sizeof(buf) - 1)
				logit(c,"http_auth: so long cnonce!");
			xnonce = buf;
		}

		authdig2(hash, xnonce, method, huri, dig);
		//logit(c,"DEBUG: http_auth: Digest: %s %s",response,dig);
		c->head.stale = 0;
		if(strcmp(response,dig) == 0) { /* response is right */
			/* 	chknonce() checks if the nonce is registered one.
			*	if it is, returns 1 ; otherwise returns 0
			*	Suppose a browser be already authenticated in a directory
			*	and then wants to enter another directory that must be
			*	authenticated. Then the browser might use old nonce
			*	because the browser does not have new nonce.
			*	-Kenar- */
			if(cnonce && chknonce(nonce,nil) && !chknonce(nonce,response)){
				ok = 1;
				putnonce(nonce,response);
			}
			else if(!cnonce && chknonce(nonce,nil)){
				ok = 1;
				rmvnonce(nonce);
			}
			else{ /* the nonce is stale or replay attack */;
				c->head.stale = 1;
				//logit(c,"DEBUG: http_auth: stale, realm=%s, nonce=%s",realm,nonce);
			}
		}
	}

R:
	/*	four types of clients will come here.
	*	0. no auth info
	*	1. basic				# passwd
	*	2. digest on RFC2069	# response
	*	3. digest on RFC2617	# response, cnonce
	*/
	if(!cnonce && !passwd){
		nextnonce = gennonce();
		putnonce(nextnonce,nil);
		c->head.nextnonce = nextnonce;
	}

	if(ok){
		authenticated = 1;
		db_put(&arealm,realm,""); /* put something into the value */
		//logit(c,"DEBUG: http_auth: ok, realm=%s, nonce=%s",realm,nonce);
		return nil;
	}
	return ret;
}

char *
gennonce(void) /* generate nonce value */
{
	char *e,*s;
	int i;
	char nonce[25];
	uchar randomkey[12];
	/*	- nonce is a random hexadecimal string of length 24 */
	s = nonce;
	e = s + sizeof(nonce);
	genrandom(randomkey, 12);
	for(i=0; i < 12; i++)
		s = seprint(s,e,"%02x", randomkey[i]);
	return strdup(nonce);
}


/*	nonce data base must be on a disk	*/
int
putnonce(char *nonce, char *res)
{
	char buf[80];
	int fd;
	if(res == nil){
		snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
		fd = create(buf,OREAD,DMDIR|0660);
		if(fd < 0)
			return -1;
		close(fd);
	}
	else{
		snprint(buf, sizeof buf,"/etc/nonce/%s/%s", nonce,res);
		fd = create(buf,OREAD,0660);
		if(fd < 0)
			return -1;
		close(fd);
	}
	return 0;
}

void
rmvnonce(char *nonce)
{
	char buf[80], buf1[80];
	int fd;
	long n, i;
	Dir *dbuf;
	snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
	fd = open(buf, OREAD);
	n = dirreadall(fd,&dbuf);
	for(i = 0; i < n; i++){
		snprint(buf1, sizeof buf1,"/etc/nonce/%s/%s", nonce);
		remove(buf1);
	}
	free(dbuf);
	remove(buf);
}


/* 	returns 0, if non existent
*	returns 1, if exist
*	-Kenar- */
int
chknonce(char *nonce, char *res)
{
	char buf[80];
	if(res == nil){
		snprint(buf, sizeof buf,"/etc/nonce/%s", nonce);
		return !access(buf,AEXIST);
	}
	snprint(buf, sizeof buf,"/etc/nonce/%s/%s", nonce, res);
	return !access(buf,AEXIST);
}
